#!/usr/bin/env python3

"""   iY Lab
  Project: Developing a tool for exploring  RNA-Protein interactions
  Name 	: Muhammad Nahin Khan
  AndrewID  : mnk1
  File Created: 01/26/2020
  Script written by M. Nahin Khan
  mnk1@andrew.cmu.edu

RNPFind is a program that allows the user to input one RNA template molecule
and to analyse the RBPs that bind on to it.

Introduction:
RNPFind takes as input the name of one RNA-encoding gene, or the coordinates of
a region of interest on the hg38 human genome. In the future we plan on
supporting an arbitrary RNA sequence as input to the program.

The program then takes as input a preferred set of data sources from which it
should extract the RNA-RBP binding information from (e.g. experimental-based
ones like POSTAR or computational-prediction-based databases like RBPDB).
Currently, RNPFind supports automated data extraction from RBPDB, ATTRACT, and
POSTAR. Furthermore, custom-data (for example data obtained from one's own
experiments) are also supported for inputting to the program. If you wish to add
support for any other data source / algorithm of your liking, I recommend
looking at "data_load_functions.py", which outlines ways for you to contribute
your own source and integrating it into RNPFind.

After loading RBP-Protein binding data, RNPFind then does internal processing to
efficiently store the binding sites of RBPs on the template RNA. For details on
how this might work, take a look at "bind_analysis.py" and
"binding_analysis_binding_sites.py".

Finally, after internal processing, RNPFind has multiple analysis features for
getting useful information out of the RBP-binding data. Some of these analysis
methods may require additional input from the user, such as certain computation
parameters. Currently, we support two main analysis features: visualization on
UCSC genome browser (within which we have integrated viewing density plots) and
correlation matrix generation. For details on supporting other analysi
functions that you may be interested in contributing, check out
"analysis_functions.py"!"""


import argparse
import os
import shutil
import sys
from datetime import datetime
from pathlib import Path

# Responsible for managing the analysis functions that manipulate the RNA-RBP
# interaction data to get a useful output:
from .analysis_functions import (
    analysis_method_functions,
    analysis_methods_supported_short,
)
from .config import (
    DEFAULT_BASE_STRINGENCY,
    RO_DATA_PATH,
    RO_DATA_TAR_NAME,
    RO_DATA_URL,
)
from .data_load_functions import data_load_sources_supported_short

# Responsible for managing the loading of RNA-RBP interaction data:
from .load_data import load_data

# Functions that help with interacting with the user to get their preference:
from .user_input import get_user_rna_preference


def rm_folder_contents(folder):
    """
    Remove all the contents in a directory.
    From: https://stackoverflow.com/a/185941/8551394

    """
    for filename in os.listdir(folder):
        file_path = os.path.join(folder, filename)
        try:
            if os.path.isfile(file_path) or os.path.islink(file_path):
                os.unlink(file_path)
            elif os.path.isdir(file_path):
                shutil.rmtree(file_path)
        except OSError:
            print("Failed to delete %s.")


import urllib.request

from tqdm import tqdm


class DownloadProgressBar(tqdm):
    def update_to(self, b=1, bsize=1, tsize=None):
        if tsize is not None:
            self.total = tsize
        self.update(b * bsize - self.n)


def download_url(url, output_path):
    with DownloadProgressBar(
        unit="B", unit_scale=True, miniters=1, desc=url.split("/")[-1]
    ) as t:
        urllib.request.urlretrieve(
            url, filename=output_path, reporthook=t.update_to
        )


def download_ro_data():
    """
    Deletes the ro-data directory and downloads the contents from
    rnpfind.com

    """
    # Delete folder
    try:
        shutil.rmtree(RO_DATA_PATH)
    except FileNotFoundError:
        pass

    # Create folder
    Path(RO_DATA_PATH).mkdir(parents=True, exist_ok=True)

    # Download tar file
    print("Downloading tar file...", file=sys.stderr)
    download_url(RO_DATA_URL, RO_DATA_TAR_NAME)

    # Extract tar file
    print("Extracting tar file (this will take some time)...", file=sys.stderr)
    shutil.unpack_archive(
        RO_DATA_TAR_NAME, extract_dir=Path(RO_DATA_PATH).parent
    )

    # Delete the tar file
    Path(RO_DATA_TAR_NAME).unlink()

    # Display confirmation of completion
    print("Done!", file=sys.stderr)


def rnpfind(
    transcript,
    sources=None,
    methods=None,
    base_stringency=None,
    out_dir=None,
    is_trackhub=False,
    is_trackhub_only=False,
):
    """
    Collect binding data of RBPs on RNA.

      :param transcript: gene name or genomic location to specify transcript
      :param sources: list of data sources to limit to for data collection
      :param methods: list of output formats to restrict to
      :param base_stringency: config option for csv output method
      :param out_dir: directory to write output files in
      :param is_trackhub: whether to generate trakchub structure
      :param is_trackhub_only: wheter to delete BED files in the end
    """

    # First, check if readonly data directory exists
    if not Path(RO_DATA_PATH).is_dir():
        # We assume that if the dir exists the data is fine; otherwise
        # the data needs to be downloaded

        # In case of corrupt data one would have to call download_ro_data()
        # manually (Or if one were to wish for just the data without analysis)
        print("Downloading data necessary for rnpfind...", file=sys.stderr)
        download_ro_data()

    # Start by getting the transcript of interest to analyze
    rna_info: dict = get_user_rna_preference(transcript)

    # what data sources does the user want to collect data from today?
    # (e.g. attract, postar, etc.)
    data_load_sources = (
        sources if sources else data_load_sources_supported_short
    )
    # print(
    #     f"Collecting data from: {', '.join(data_load_sources)}",
    #     file=sys.stderr,
    # )
    # load RNA-RBP interaction data using the selected data sources on the RNA
    # molecule of interest big_storage stores data on binding sites of RBPs on
    # the RNA molecule from each data source. For more details on how
    # big_storage is structured, consult load_data.py!
    big_storage = load_data(data_load_sources, rna_info)

    # BIOGRID is a database that stores information on protein-protein
    # interaction evidence in the literature from experiment. In future versions
    # of RNPFind, BIOGRID data should be helpful in a variety of inquiries when
    # investigating relationships between RBPs that bind on RNA molecules.

    # Todo: Consider collecting and using BIOGRID data in a meaningful way

    # Give a quick summary on the data that has been loaded, in terms of total
    # number of RBPs and binding sites collected for the RNA molecule of
    # interest
    no_rbps = 0
    no_sites = 0
    for no_rbp, no_site in [big_storage[k].summary() for k in big_storage]:
        no_rbps += no_rbp
        no_sites += no_site

    print(
        f"Collected data for {no_rbps} RBPs with {no_sites} binding sites",
        file=sys.stderr,
    )

    # We now proceed to perform any number of analysis methods that the user
    # may wish to apply to the data obtained
    analysis_methods = methods if methods else analysis_methods_supported_short
    # print(
    #     "Producing the following output formats:"
    #     f" {', '.join(analysis_methods)}",
    #     file=sys.stderr,
    # )

    if out_dir:
        # The user specified an out directory
        # Create it if it does not exist
        # TODO: consider error handling
        Path(out_dir).mkdir(parents=True, exist_ok=True)
    else:
        # The user did not specify an out directory
        default_path = Path.cwd() / rna_info["official_name"].lower()

        while default_path.is_dir():
            # The default path already exists, append the folder name with
            # current date and time
            time_list = datetime.now().timetuple()
            time_list = [str(x) for x in time_list]
            time_date = "-".join(time_list[0:6])  # year to seconds
            default_path = default_path.parent / (
                f"{default_path.name}-{time_date}"
            )

        default_path.mkdir(parents=True)
        out_dir = str(default_path)

    rm_folder_contents(out_dir)

    for analysis_method in analysis_methods:
        print(f"Generating {analysis_method}", file=sys.stderr)
        analysis_method_function = analysis_method_functions[analysis_method]

        configs = {"out_dir": out_dir}
        if analysis_method == "csv":
            configs["base_stringency"] = (
                base_stringency if base_stringency else DEFAULT_BASE_STRINGENCY
            )
        if analysis_method == "bed":
            configs["trackhub"] = is_trackhub
            configs["trackhub-only"] = is_trackhub_only

        analysis_method_function(big_storage, rna_info, configs=configs)

    print("Done!", file=sys.stderr)


def main():
    """
    main function responsible for parsing commandline args

    """
    out_formats = analysis_method_functions.keys()

    parser = argparse.ArgumentParser(
        add_help=False,
        description="Get binding sites of RBPs on a given transcript",
    )
    parser.add_argument(
        "transcript",
        help="Specify with the name of a gene (e.g. 'Malat1')"
        " or as hg38 chromosome coordinates given as"
        " <chr_no>:<start_coord>-<end_coord> (e.g. 5:4000-14000)"
        ". Note that chromosome number is X, Y, M(T), or a number between"
        " 1 and 22",
    )
    parser.add_argument(
        "-h",
        "--help",
        action="help",
        default=argparse.SUPPRESS,
        help="Show this help message and exit.",
    )
    parser.add_argument(
        "-s",
        "--sources",
        choices=data_load_sources_supported_short,
        nargs="+",
        help="Pick source(s) for RBP binding data. Pick any non-empty subset"
        f" of values from {{{', '.join(data_load_sources_supported_short)}}}."
        " If unspecified, all sources are selected.",
        metavar=("<source 1>", "<source 2>"),
    )
    parser.add_argument(
        "-o",
        "--out-dir",
        metavar="<dir>",
        help="Directory to write output files in. If unspecified, a folder is"
        " created and written to (if possible)",
    )
    parser.add_argument(
        "-f",
        "--out-format",
        choices=out_formats,
        nargs="+",
        metavar=("<format 1>", "<format 2>"),
        help="Choose format(s) to store binding site data in. Pick any"
        f" non-empty subset from {{{', '.join(out_formats)}}}."
        " If unspecified, all formats are created in the output directory.",
    )
    parser.add_argument(
        "-b",
        "--base-stringency",
        type=int,
        help="The number of bases beween two RBP-binding-sites before they"
        " are considered to be competing (used only for csv output format)."
        f" The default value is {DEFAULT_BASE_STRINGENCY}.",
        metavar="<N>",
        default=DEFAULT_BASE_STRINGENCY,
    )
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "--trackhub",
        action="store_true",
        help="If specified, generates a trackhub structure using bed files"
        " (only used for 'bed' output format)",
        default=False,
    )
    group.add_argument(
        "--trackhub-only",
        action="store_true",
        help="If specified, generates a trackhub structure using bed files and"
        " deletes all original bed files (only used for 'bed' output format)",
        default=False,
    )

    args = parser.parse_args()
    rnpfind(
        args.transcript,
        args.sources,
        args.out_format,
        args.base_stringency,
        args.out_dir,
        args.trackhub,
        args.trackhub_only,
    )


if __name__ == "__main__":
    main()



"""
Tests the RNPFind analysis functions module for internal consistency.

"""
import inspect
import unittest

from src.rnpfind.analysis_functions import (
    analysis_method_functions,
    analysis_methods_supported_long,
    analysis_methods_supported_short,
)


class TestAnalysisMethods(unittest.TestCase):
    """
    Tests to ensure that the analysis method related variables are properly set
    up.

    """

    def test_equal_number_of_methods(self):
        """
        Checks that the number of analysis functions are equal across variables
        describing them.

        """
        self.assertEqual(
            len(analysis_methods_supported_short),
            len(analysis_methods_supported_long),
        )
        self.assertEqual(
            len(analysis_method_functions),
            len(analysis_methods_supported_short),
        )

    def test_appropriate_name(self):
        """
        Checks that the analysis functions are given appropriate 'short' names.
        """
        for short_name, long_name in zip(
            analysis_methods_supported_short, analysis_methods_supported_long
        ):
            for token in short_name.split("_"):
                self.assertIn(token.lower(), long_name.lower())

    def test_method_dict_keys(self):
        """
        Checks whether analysis_method_functions describes the functions listed
        in analysis_methods_supported_short.

        """

        # equality of size from a previous test along with injection here
        # assures bijection.
        for method in analysis_methods_supported_short:
            self.assertIn(method, analysis_method_functions)

    def test_appropriate_method_type(self):
        """
        Checks that each of the analysis methods have a consistent type
        signature (only enforced for input; return type is different depending
        on analysis type)

        """

        for method in analysis_method_functions.values():
            self.assertEqual(
                list(inspect.signature(method).parameters)[0], "big_storage"
            )

            self.assertEqual(
                list(inspect.signature(method).parameters)[1], "rna_info"
            )


if __name__ == "__main__":
    unittest.main()



"""
Tests the user input functions module for correctness.

"""
import unittest

from src.rnpfind.gene_coordinates import Chromosome
from src.rnpfind.user_input import parse_genome_coord


class TestParseGenomeCoord(unittest.TestCase):
    """
    Tests to ensure that parsing of genome coordinates is correct

    Note that this does not test the ability of the parser to detect incorrect
    chromsome number based on size of chromosome

    """

    def test_normal_success(self):
        """
        Checks whether normal coordinates are validated correctly

        """

        case_1 = "3:400-14000"
        case_2 = "1:360-5000"
        case_3 = "22:3456-23456"
        result_1 = parse_genome_coord(case_1)
        result_2 = parse_genome_coord(case_2)
        result_3 = parse_genome_coord(case_3)

        self.assertTrue(result_1["success"])
        self.assertTrue(result_2["success"])
        self.assertTrue(result_3["success"])

        self.assertEqual(result_1["chr_n"], Chromosome(3))
        self.assertEqual(result_2["chr_n"], Chromosome(1))
        self.assertEqual(result_3["chr_n"], Chromosome(22))

        self.assertEqual(result_1["start_coord"], 400)
        self.assertEqual(result_2["start_coord"], 360)
        self.assertEqual(result_3["start_coord"], 3456)

        self.assertEqual(result_1["end_coord"], 14000)
        self.assertEqual(result_2["end_coord"], 5000)
        self.assertEqual(result_3["end_coord"], 23456)

    def test_invalid_chr_no(self):
        """
        Check whether genomic coordinate outside valid range fails

        """
        case_1 = "24:100-1000"
        case_2 = "0:130-2345"
        case_3 = "-1:2345-7653"
        case_4 = "23:2345-234546"

        result_1 = parse_genome_coord(case_1)
        result_2 = parse_genome_coord(case_2)
        result_3 = parse_genome_coord(case_3)
        result_4 = parse_genome_coord(case_4)

        self.assertFalse(result_1["success"])
        self.assertFalse(result_2["success"])
        self.assertFalse(result_3["success"])
        self.assertFalse(result_4["success"])

    def test_invalid_range(self):
        """
        Check that wrong coordinate range fails

        """

        case_1 = "2:5000-500"
        case_2 = "4:6000-5800"
        case_3 = "X:14000-4000"

        result_1 = parse_genome_coord(case_1)
        result_2 = parse_genome_coord(case_2)
        result_3 = parse_genome_coord(case_3)

        self.assertFalse(result_1["success"])
        self.assertFalse(result_2["success"])
        self.assertFalse(result_3["success"])

    def test_bad_start_coord(self):
        """
        Check that bad start coord results in fail-outut

        """
        case_1 = "2:abd-50000"
        case_2 = "4:cyz-580000"
        case_3 = "X:hgff-4000000"

        result_1 = parse_genome_coord(case_1)
        result_2 = parse_genome_coord(case_2)
        result_3 = parse_genome_coord(case_3)

        self.assertFalse(result_1["success"])
        self.assertFalse(result_2["success"])
        self.assertFalse(result_3["success"])

    def test_bad_end_coord(self):
        """
        Check that bad end coord results in fail-outut

        """
        case_1 = "2:500-xyz"
        case_2 = "4:4567-tac"
        case_3 = "X:617-rbp"

        result_1 = parse_genome_coord(case_1)
        result_2 = parse_genome_coord(case_2)
        result_3 = parse_genome_coord(case_3)

        self.assertFalse(result_1["success"])
        self.assertFalse(result_2["success"])
        self.assertFalse(result_3["success"])

    def test_negative_coord(self):
        """
        Check whether parsing negative coordinates fails

        """
        case_1 = "12:-234:-54"
        case_2 = "3:-234:54"
        case_3 = "X:450--600"

        result_1 = parse_genome_coord(case_1)
        result_2 = parse_genome_coord(case_2)
        result_3 = parse_genome_coord(case_3)

        self.assertFalse(result_1["success"])
        self.assertFalse(result_2["success"])
        self.assertFalse(result_3["success"])

    def test_letter_success(self):
        """
        Checks whether genomic coordinates with letters parse correctly

        """
        case_1 = "M:4000-15000"
        case_2 = "M:400000-15000000"
        case_3 = "MT:4000-15000"
        case_4 = "MT:400000-15000000"
        case_5 = "X:100-300"
        case_6 = "Y:300-4000"

        result_1 = parse_genome_coord(case_1)
        result_2 = parse_genome_coord(case_2)
        result_3 = parse_genome_coord(case_3)
        result_4 = parse_genome_coord(case_4)
        result_5 = parse_genome_coord(case_5)
        result_6 = parse_genome_coord(case_6)

        self.assertTrue(result_1["success"])
        self.assertTrue(result_2["success"])
        self.assertTrue(result_3["success"])
        self.assertTrue(result_4["success"])
        self.assertTrue(result_5["success"])
        self.assertTrue(result_6["success"])

        self.assertEqual(result_1["chr_n"], Chromosome("M"))
        self.assertEqual(result_2["chr_n"], Chromosome("M"))
        self.assertEqual(result_3["chr_n"], Chromosome("M"))
        self.assertEqual(result_4["chr_n"], Chromosome("M"))
        self.assertEqual(result_5["chr_n"], Chromosome("X"))
        self.assertEqual(result_6["chr_n"], Chromosome("Y"))

        self.assertEqual(result_1["start_coord"], 4000)
        self.assertEqual(result_2["start_coord"], 400000)
        self.assertEqual(result_3["start_coord"], 4000)
        self.assertEqual(result_4["start_coord"], 400000)
        self.assertEqual(result_5["start_coord"], 100)
        self.assertEqual(result_6["start_coord"], 300)

        self.assertEqual(result_1["end_coord"], 15000)
        self.assertEqual(result_2["end_coord"], 15000000)
        self.assertEqual(result_3["end_coord"], 15000)
        self.assertEqual(result_4["end_coord"], 15000000)
        self.assertEqual(result_5["end_coord"], 300)
        self.assertEqual(result_6["end_coord"], 4000)

    def test_gibberish(self):
        """
        Check that random gibberish fails

        """
        case_1 = "gwfefrgr"
        case_2 = "23:234:23:123"

        result_1 = parse_genome_coord(case_1)
        result_2 = parse_genome_coord(case_2)

        self.assertFalse(result_1["success"])
        self.assertFalse(result_2["success"])


if __name__ == "__main__":
    unittest.main()



"""
Defines the ucsc_visualize function.
This function takes care of generating binding sites, converting them to
appropriate formats (BED files, then bigBED files, wig, bigwig, etc.),
generating density plots, setting them up as a track hub structure, and finally
uploading them to a server online, so that the binding sites can be visualized.
"""

import os
import shutil
import sys
from pathlib import Path

from .populate_rbp_binding_sites_script import populate_binding_sites
from .populate_trackhub import (
    convert_bed_to_bb,
    convert_wig_to_bw,
    density_plot,
    populate_local_track_hub,
)


def ucsc_visualize(big_storage, rna_info, configs=None):
    """
    This function takes care of generating binding sites, converting them to
    appropriate formats (BED files, then bigBED files, wig, bigwig, etc.),
    generating density plots, setting them up as a track hub structure, and
    finally uploading them to a server online, so that the binding sites can be
    visualized.

    The function returns a URL where the binding sites may be viewed.

    :param big_storage: the Storage instance containing the binding sites
    :param rna_info: a dictionary specifying relevant RNA information, such as
                     the name of the RNA molecule, its location in the genome,
                     etc.
    :param print: if specified, allows for printputs to be redirected.
                All status updates are directed to the "print" function, if it is
                specified.
                (Default value = None)
    :param total_steps: don't use, will be deprecated soon.
                        (Default value = 7)
    :param configs: gives additional configurations. Supported keys include:
        - out_dir: the directory to write files to

    """

    data_load_sources = big_storage.keys()

    overarching_path = str(Path(configs["out_dir"]) / "bed")
    populate_binding_sites(
        big_storage, rna_info, data_load_sources, overarching_path
    )

    if not configs["trackhub"] and not configs["trackhub-only"]:
        # Trackhub is not needed at all...
        return

    convert_bed_to_bb(overarching_path, data_load_sources)

    rbp_no_dict = density_plot(
        big_storage, rna_info, data_load_sources, overarching_path
    )

    convert_wig_to_bw(overarching_path, data_load_sources)

    print("Generating trackhub structure...", file=sys.stderr)

    local_dir = Path(overarching_path).parent / "trackhub"

    rbp_peaks = {
        k: max(big_storage[k].sum_over_all().return_depth())
        for k in big_storage
    }

    populate_local_track_hub(
        overarching_path, rna_info, local_dir, rbp_no_dict, rbp_peaks
    )

    # Resolving links
    shutil.copytree(local_dir, f"{local_dir}-copy")
    # os.system(f"cp -rL {local_dir} {local_dir}-copy")
    os.system(f"rm -rf {local_dir}")
    os.system(f"mv {local_dir}-copy {local_dir}")

    os.system("find . -name '*.bb' | xargs rm")

    print(f"Created directory {local_dir}...", file=sys.stderr)

    if configs["trackhub-only"]:
        print("Removing original bed files...", file=sys.stderr)
        os.system(f"rm -rf {overarching_path}")

    # print("The link to your hub has been generated:")

    # hub_url = (
    #     "https://rnpfind.com/static/ucsc-tracks/"
    #     + date_time_folder_name
    #     + "/"
    #     + hub_name.replace(" ", "+")
    #     + ".hub.txt"
    # )

    # ucsc_url = (
    #     "https://genome.ucsc.edu/cgi-bin/hgTracks?db="
    #     + GENOME_VERSION
    #     + "&hubUrl="
    #     + hub_url
    #     + "&position=chr"
    #     + str(rna_info["chr_n"])
    #     + ":"
    #     + str(rna_info["start_coord"])
    #     + "-"
    #     + str(rna_info["end_coord"])
    # )

    # print(ucsc_url)
    # return ucsc_url



"""
This module defines a data load function. In particular, it loads a custom,
user-specified data that is available in custom_binding_data.py
"""
from .custom_binding_data import custom_data


def custom_data_load(rna_info):
    """
    Loads custom binding data specified in custom_binding_data.py and returns
    a Generator / Iterator object with the binding sites.

    :param rna_info: A dictionary containing the name of the RNA molecule,
        specified under its 'official_name' key.

    """
    rna = rna_info["official_name"]
    for rbp, binding_sites in custom_data[rna].items():
        for start, end in binding_sites:
            annotation = ""
            yield rbp, start, end, annotation



"""
An old file that I did not integrate into RNPFind. But maybe it has potential?
(probably not to anyone else)
"""

# from bind_analysis import BindingSites
from operator import itemgetter

# from merge_annotation_funcs import generate_merge_func
# import statistics  # Standard deviation mostly
# import operator

firstTwoItems = itemgetter(0, 1)
firstItem = itemgetter(0)
secondItem = itemgetter(1)
thirdItem = itemgetter(2)


def per_binding_site_analysis(big_storage, rna_info):
    """
    An analysis function that generates information on each binding site bound
    to on an RNA of interest by an RBP of interest. Information for each binding
    site includes the binding sites of other RBPs, competitive/cooperative
    relations, etc.

    :param big_storage: a dictionary indexed by data sources that correspond to
        Storage instances containing binding site data extracted from said data
        source.
    :param rna_info: A dictionary containing data on RNA of interest, such as
        its name and location on the human genome.

    """
    # This function previously required these inputs
    # bigStorage, AUF1Filter, analysis_per_binding_site_window,
    # analysis_per_binding_site_competitive_range, analysis_sources

    # Variable that comes useful later.
    print("unimplemented for now!")
    print(big_storage, rna_info, "", "")

    # TODO:This function needs to be reconsidered with respect to RNPFind and
    # generalized for one template

    # binding_info = {}
    #
    # # CHECK IF THIS SELECTION IS IMPORTANT OR NOT
    # # Is this selection important??
    # # storageSpace = select(bigStorage,analysis_sources)
    #
    # # For each lncRNA, we want to analyze the AUF1 binding sites
    # lncRNA = RNA
    # lncRNA_storage = storageSpace
    #
    # print("")
    # print("*****************")
    # print("analysing AUF1 sites on", lncRNA + "...\n")
    # print("*****************")
    #
    # # Get the RBPs that bind close to sites on AUF1
    # filtered_storage_dict = lncRNA_storage.sites_analysis(
    #     "AUF1", bp_threshold=analysis_per_binding_site_window
    # )

    # # Sort AUF1 sites by readCount
    # auf1_sites = sorted(filtered_storage_dict.keys(),
    #                     key=lambda k: int(k[2].split()[1]), reverse=True)
    #
    # # Variables for later analysis
    # distance_read_collection = {}
    #
    # # Analyse each site
    # for site in auf1_sites:
    #
    #     # Number of times the binding site was detected in RNASeq
    #     readCount = int(site[2].split()[1])
    #
    #     # Get collection of RBPs and their binding sites close to the current
    #     # AUF1 binding site being considered
    #     dB_storage = filtered_storage_dict[site]
    #
    # print(
    #     "\nanalysing binding site from",
    #     str(site[0]) + "bps to " + str(site[1]) + "bps; readCount =",
    #     readCount, 'from', site[2].split()[0]
    # )

    #     # We'll collect data in a list then sort them later.
    #     listOfThingsToPrint = []
    #
    #     # For each RBP and its binding sites:
    # for rbp, binding_sites in dB_storage.items():
    #     # Get the distances away from the current AUF1 site being considered
    #     distances = list(
    #         map(lambda s: BindingSites.distance(s, site), binding_sites)
    #     )

    #     # Separate binding sites and annotation:
    #     closest_sites = list(map(firstTwoItems, binding_sites))
    #     annotations = list(map(thirdItem, binding_sites))

    #     # Check if the RBP has experimental evidence for binding AUF1:
    #     isBindAUF1 = AUF1Filter(rbp)

    #     listOfThingsToPrint.append(
    #         (rbp, closest_sites, distances, annotations, isBindAUF1))

    #     # Variables to analyse later
    #     distance_read_collection[rbp] = (
    #         distance_read_collection.get(rbp, [])
    #         + list(map(lambda d: (d, readCount), distances))
    #     )

    # # Now sort and print!
    # for toPrint in sorted(listOfThingsToPrint,
    #                       key=lambda k: min(k[2])):

    #     # Unpack
    #     rbp, closest_sites, distances, annotations, isBindAUF1 = toPrint

    #     # Modify
    #     closest_sites = str(closest_sites)[1:-1]
    #     isBindAUF1 = 'yes' if isBindAUF1 else 'no'
    #     isAllCompetitive = (
    #         all(map(
    #         lambda k: k <= analysis_per_binding_site_competititive_range,
    #         distances
    #         ))
    #     )

    #     isAllCooperative = (
    #         all(map(
    #         lambda k: k >= analysis_per_binding_site_competititive_range,
    #         distances))
    #     )
    #     isAllCooperativeORCompetitive = (
    #         'competitive' if isAllCompetitive else
    #         ('cooperative' if isAllCooperative else 'neither')
    #     )
    #     distances = str(distances)[1:-1]

    #     # TODO: Consider methods of tracking datasources despite multiple
    #     # source merges

    #     data_load_source = analysis_sources[0]
    #     merge_func = generate_merge_func(data_load_source)

    #     annotations = merge_func(annotations)

    #     # Repack
    #     toPrint = (
    #         rbp, closest_sites, distances, annotations, isBindAUF1,
    #         isAllCooperativeORCompetitive
    #     )

    #         if False:
    #             # Print
    #             print(";".join(map(str, toPrint)))
    #
    #     print("*****************")
    #     print("*****************")
    #     print("*****************")
    #
    #     if True:
    #         info_collection = {}
    #
    #         def sorter(item):
    #             rbp = item[0]
    #             dist_read = item[1]
    #             flat_read_list = []
    #             for dist, read in dist_read:
    #                 flat_read_list += [dist] * read
    #
    #             reads = list(map(secondItem, dist_read))
    #             deviation = (
    #                 statistics.stdev(flat_read_list) if len(reads) > 1 else 100
    #             )
    #             return (
    #                 -sum(reads) / deviation if deviation != 0 else -sum(reads)
    #             )

    #             # return (-sum(reads), deviation)

    #         for rbp, dist_read in sorted(
    #             distance_read_collection.items(), key=sorter
    #         ):

    #             # Only because weighted sum standard deviation is difficult to
    #             # calculate:
    #             flat_read_list = []
    #             for dist, read in dist_read:
    #                 flat_read_list += [dist] * read

    #             reads = list(map(secondItem, dist_read))
    #             distance = list(map(firstItem, dist_read))
    #             deviation = (
    #                 statistics.stdev(flat_read_list) if len(reads) > 1 else 100
    #             )

    #             info_collection[rbp] = (
    #                 rbp, statistics.median(flat_read_list), deviation,
    #                 sum(reads), len(set(reads)), len(reads), min(distance),
    #                 max(distance), ['no', 'yes'][AUF1Filter(rbp)],
    #                 str(dist_read)[1:-1][:100]
    #             )

    #             if True:
    #                 print(";".join(map(str, info_collection[rbp])))

    #     binding_info[lncRNA] = (filtered_storage_dict, auf1_sites,
    #                             distance_read_collection, info_collection)

    # countofInvolvement = {}
    # for lncRNA in ["Neat1", "Malat1"]:
    #     distance_read_collection = binding_info[lncRNA][2]
    #     for rbp, dist_read in distance_read_collection.items():
    #         competitiveScore = (
    #             sum(map(
    #             lambda t: (
    #                 t[1] if t[0] < analysis_per_binding_site_competititive_range
    #                 else 0
    #             ),
    #             dist_read
    #             ))
    #         )

    #         cooperativeScore = (
    #             sum(map(
    #             lambda t: (
    #                 t[1] if t[0] > analysis_per_binding_site_competititive_range
    #                 else 0
    #             ),
    #             dist_read
    #             ))
    #         )

    #         if rbp not in countofInvolvement: countofInvolvement[rbp] = [
    #             0, 0, 0, 0
    #         ]

    #         if lncRNA == "Malat1":
    #             countofInvolvement[rbp][0] += competitiveScore
    #             countofInvolvement[rbp][1] += cooperativeScore
    #         elif lncRNA == "Neat1":
    #             countofInvolvement[rbp][2] += competitiveScore
    #             countofInvolvement[rbp][3] += cooperativeScore
    #         else:
    #             raise ValueError('Fatal Error, please debug')

    # # normalization = [0]*4
    # # for scores in countofInvolvement.values():
    # # 	for j in [0,1,2,3]:
    # # 		normalization[j] = max(scores[j],normalization[j])

    # total_neat1_reads = (
    #     sum(map(lambda k: int(k[2].split()[1]), storageSpace['Neat1']["AUF1"]))
    # )
    # total_malat1_reads = (
    #     sum(map(lambda k: int(k[2].split()[1]), storageSpace['Malat1']["AUF1"]))
    # )
    # normalization = [total_malat1_reads, total_malat1_reads,
    #                  total_neat1_reads, total_neat1_reads]
    # print(normalization, "-->normalization")

    # def sorter(item):
    #     rbp = item[0]
    #     scores = item[1]
    #     norm_scores = list(map(operator.truediv, scores, normalization))
    # What should I prioritize?
    # scores [a , b, c, d]
    # 		 ^   ^  ^  ^
    # 		 |   |  |  |
    # 		/   /    \  \
    # 	Malat1 		  Neat1
    # comp., cooper     comp., cooperative...

    # Presumably, Malat1 and Neat1 have a similar level
    # of binding to AUF1... In that case, perhaps there isn't
    #     # as much competitive inhibition, and the thing that
    #     # destabiilzes Neat1 but not Malat1 is a cooperative helper
    #     # that's with Neat1 but not on Malat1.
    #     # Alternatively, if this presumption is wrong, we can expect
    #     # a competitive RBP to be affecting on Malat1 but not on
    #     # Neat1.
    #     # We conclude two ways to evaluate interest:
    #     #	1. Maximize (Cooperation on Neat1 - Cooperation on Malat1)
    #     #	2. Maximize (Competitive on Malat1 - Competitive on Neat1)
    #
    #     # For now, i've chosen to maximize both at the same time:
    #
    #     comp_malat1 = norm_scores[0]
    #     coOP_malat1 = norm_scores[1]
    #     comp_neat1 = norm_scores[2]
    #     coOP_neat1 = norm_scores[3]
    # return max(abs(coOP_neat1 - coOP_malat1), abs(comp_malat1 - comp_neat1))
    #
    # for rbp, scores in sorted(countofInvolvement.items(),
    #                           key=sorter, reverse=True):
    #     norm_scores = list(map(operator.truediv, scores, normalization))
    #     item = ([rbp] + scores + ['yes' if AUF1Filter(rbp) else 'no'] +
    #             norm_scores)
    #     print(";".join(map(str, item)))



"""
Defines the class Storage.
Storage stores a collection of BindingSites together. It allows for convenient
operations such as looking up which BindingSites instances lie near or in a
given interval, calculating correlations between the BindingSites instances,
etc.

Storage is used in RNPFind to represent one RNA molecule, to which multiple
differet RNA binding proteins (RBPs) can bind. Thus each BindingSite represents
an RBP molecule, within a Storage.
"""

import difflib  # Just to suggest keys when mis-spelt!
from operator import itemgetter

from .binding_analysis_binding_sites import BindingSites

firstItem = itemgetter(0)
secondItem = itemgetter(1)
thirdItem = itemgetter(2)


# To-do:
# Consider the need for synonym_function as an initialization parameter, and
# think about membership of genes when genes are represented as comma separated
# values sometimes.

# Consider the __contains__() function.
# Priority: low
# Justification: most people will iterate through the storage instead of testing
# with 'in'.


class Storage:
    """
    Storage stores a collection of BindingSites together. It allows for
    convenient operations such as looking up which BindingSites instances lie
    near or in a given interval, calculating correlations between the
    BindingSites instances, etc.

    Storage is used in RNPFind to represent one RNA molecule, to which multiple
    differet RNA binding proteins (RBPs) can bind. Thus each BindingSite
    represents an RBP molecule, within a Storage.

    This class stores data about a lncRNA and RBPs that might bind to it, as
    well as the sites at which they bind (using dictionaries that store
    BindingSites values)
    """

    def __init__(self, synonym_func=-1, annotation_merge_func=-1):

        # Stores RBP names as keys and Binding Sites as values
        self._rbps = {}

        # When self_analysis is called, a matrix like table (but actually
        # just a nested dictionary) is created and stored here. For any two
        # RBPs x and y, self.corr_table[x][y] gives the correlation between
        # x and y's binding sites (when y is "thrown" at x)
        self.corr_table = -1

        # The same correlation table aforementioned, but [x][y] and [y][x]
        # give the same result, which is the f-score (2pq/(p+q) where p and
        # q are the scores from [x][y] and [y][x]). This is important I think
        # because some RBPs seem to bind almost everywhere, so correlation is
        # a one-way street
        self.corr_table_f_measure = -1

        # Stores a list of tuples sorted by the distance scores. The tuples
        # are of the form (x,y,score) where x and y are RBP gene names. Note
        # that the scores here are actually the f-scores!
        self.corr_sorted = -1

        # When the analysis was done, what was the base pair stringency used?
        self.corr_bp_threshold = -1

        # A function that defines how binding site annotations are merged.
        self.merge = annotation_merge_func

        # An init argument, synonym_dict, is an optional argument that should be
        # a function mapping gene synonyms to the official symbol, in case
        # multiple data sources have the same gene referred to using different
        # symbols.
        if synonym_func == -1:
            self.synonym_func = lambda x: x
        else:
            self.synonym_func = synonym_func

    def __repr__(self, display_meta=False):
        rbp_reprs = []
        for rbp, binding_sites in self._rbps.items():
            site_str = binding_sites.__repr__(display_meta=display_meta)
            i = site_str[50:].find("),")
            rbp_reprs += [rbp + ": " + site_str[0 : 52 + i] + "..."]

        if len(rbp_reprs) > 20:
            rbp_reprs = rbp_reprs[:8] + ["..."] * 2 + rbp_reprs[-8:]
        return (
            "\nA storage variable containing binding sites for "
            + "the following "
            + str(len(self))
            + " RBPs: \n\n"
            + "\n".join(rbp_reprs)
        )

    def __str__(self):
        return self.__repr__(display_meta=True)

    def __len__(self):
        return self._rbps.__len__()

    def __iter__(self):
        return self._rbps.__iter__()

    def get_rbps(self):
        """
        returns a list of RBPs (BindingSite ID's) that are binding to the
        Storage instance (the RNA molecule instance)
        """
        return self._rbps.keys()

    def items(self):
        """
        returns a list of (RBP, binding sites) tuples that are bound on this
        RNA molecule.

        Another view is: returns a list of (BindingSite ID's, BindingSites) that
        the Storage instance contains
        """
        return self._rbps.items()

    def values(self):
        """
        Returns a list of BindingSites instances that are involves with the
        Storage instance (without listing the RBP's responsible for the binding
        sites)
        """
        return self._rbps.values()

    def __getitem__(self, item):
        if isinstance(item, str):
            item = item.upper()
            item = item.strip()
            if item in self._rbps:
                return self._rbps[item]

            possible_list = difflib.get_close_matches(item, self._rbps.keys())

            if possible_list:
                print("No such key found... Did you mean:")
                print(possible_list)

            else:
                print("sorry wrong key:", item)

            raise KeyError(item)

        try:
            # treat item as a list of rbps and get the subset-slice out to
            # the caller
            subset_sites = Storage()
            for rbp in item:
                subset_sites[rbp] = self[rbp]
            return subset_sites

        except TypeError as err:
            print("You indexed with type:", type(item))
            raise ValueError(
                "Please index with iterables containing strings"
            ) from err

    def corr_reset(self):
        """
        Resets correlation data stored internally.
        In case new data is added, it's a good idea to reset correlation data
        """
        self.corr_table = -1
        self.corr_table_f_measure = -1
        self.corr_sorted = -1

    def __setitem__(self, item, value):
        # Discourage this use mostly, but if the types are right why not

        if isinstance(item, str) and isinstance(value, BindingSites):
            item = item.upper()
            item = item.strip()
            self._rbps[item] = value
            self.corr_reset()

        else:
            raise ValueError("Assign gene names to BindingSites please")

    def summary(self):
        """
        Allows user to inspect the Storage instance.

        Show binding site information in sorted order:
        Example usage: neat1_storage.summary()

        """

        summary_info = []
        for rbp in self._rbps:
            summary_info.append((rbp, len(self._rbps[rbp])))

        return len(self._rbps), sum([k for (a, k) in summary_info])

    def self_analysis(
        self,
        bp_threshold=30,
        display_threshold=0.8,
        verbose=False,
        progress_feedback=True,
    ):
        """
        In case you are looking for fun and want to do a correlation study
        between all the RBPs pairwise on the lncRNA you are studying.

        Especially useful if you have real binding data in my opinion.

        :param bp_threshold: number of bases beyond which two RBPs are
                             considered too far from each other (in terms of
                             location of their binding sites)
                             (Default value = 30)
        :param display_threshold: Threshold correlation value below which the
                                  pairwise RBP correlations are not shown.
                                  (Default value = 0.8)
        :param verbose: specifies level of detail in output
                        (Default value = False)
        :param progress_feedback: specifies whether the progress of analysis is
                                  printed continuously as the correlations are
                                  calculated
                                  (Default value = True)

        """

        # If analysis hasn't happened yet or different stringency is being used:
        if self.corr_table == -1 or self.corr_bp_threshold != bp_threshold:
            corr_table = {}  # Stores the correlation table (nested dictionary)
            prev_num = 0
            num_j = 0
            for num_i, i in enumerate(self._rbps):
                for num_j, j in enumerate(self._rbps):
                    if i not in corr_table:
                        corr_table[i] = {}
                    corr_table[i][j] = self._rbps[i].dist(
                        self._rbps[j], bp_threshold
                    )

                # Progress should be printed since this can take some time
                percentage = (
                    (num_i * len(self._rbps) + num_j)
                    / len(self._rbps) ** 2
                    / 0.01
                )
                if progress_feedback:
                    if (
                        round(percentage) % 20 == 0
                        and round(percentage) != prev_num
                    ):
                        print(round(percentage), "%" + "complete")
                        prev_num = percentage

            # Stores the same scores as corr_table above but f-scores instead
            corr_table_f_score = {}

            tuple_list = []  # Keeps it in a tuple form for easy sorting
            for num_i, i in enumerate(self._rbps):
                for num_j, j in enumerate(self._rbps):
                    score_1 = corr_table[i][j]
                    score_2 = corr_table[j][i]

                    if i not in corr_table_f_score:
                        corr_table_f_score[i] = {}

                    if score_1 == 0 and score_2 == 0:
                        corr_table_f_score[i][j] = 0
                    else:
                        corr_table_f_score[i][j] = (
                            2 * score_1 * score_2 / (score_1 + score_2)
                        )

                    tuple_list.append((i, j, corr_table_f_score[i][j]))

            sorted_tuple_list = sorted(tuple_list, key=thirdItem, reverse=True)

            self.corr_table = corr_table
            self.corr_table_f_measure = corr_table_f_score
            self.corr_sorted = sorted_tuple_list
            self.corr_bp_threshold = bp_threshold

            if verbose:
                print("Data was saved in storage")

        if verbose:
            print(
                "Some of the highest score pairs above threshold"
                " (0.8 by default):"
            )

            for output in self.corr_sorted:  # sorted three element tuple list
                if 1.0 > output[2] > display_threshold:
                    print(output)

        return (
            self.corr_table,
            self.corr_table_f_measure,
            self.corr_sorted,
            self.corr_bp_threshold,
        )

    def lookup(
        self, first_rbp, second_rbp=-1, display_threshold=-0.1, bp_threshold=30
    ):
        """
        gets the correlation value for a specific RBP and all RBPs or a specific
        RBP and another one.
        :param first_rbp: the first RBP
        :param second_rbp: the second RBP, or -1 if you need against all
                  (Default value = -1)
        :param display_threshold:  the cutoff value below which correlation
                                   pairs should not be shown
                                   (Default value = -0.1)
        :param bp_threshold:  the number of bases beyond which two RBPs are
                              considered too far in terms of the distance
                              between a pair of their binding sites
                              (Default value = 30)

        """
        if self.corr_table == -1 or self.corr_bp_threshold != bp_threshold:
            self.self_analysis(
                bp_threshold=bp_threshold, display_threshold=1.1
            )

        if second_rbp == -1:
            to_return_list = []
            corr_entries = self.corr_table_f_measure.get(first_rbp, {})
            for key in sorted(
                corr_entries, key=corr_entries.get, reverse=True
            ):
                if float(corr_entries[key]) > display_threshold:
                    to_return_list.append((key, corr_entries[key]))

            return to_return_list

        return self.corr_table_f_measure.get(first_rbp, {}).get(second_rbp, 0)

    def lookup_table(self):
        """
        Returns the correlation table for this RNA molecule.
        A correlation table describes the extent to which the RBPs binding to
        this molecule have similar binding patterns on this RNA.
        """
        return self.corr_table_f_measure

    def binds_near(self, interval_range, bp_threshold=30):
        """
        This function takes a tuple representing an interval that one wants
        to test on the lncRNA and returns a Storage of RBPs binding to or near
        that interval

        :param interval_range: in the form (start, end) or (start, end, _)
        :param bp_threshold:  (Default value = 30)

        """
        start, end, *_ = interval_range
        extended_range = (start - bp_threshold, end + bp_threshold)

        to_return = Storage()

        for k in self._rbps:
            if self._rbps[k].is_overlap(extended_range):
                to_return[k] = self._rbps[k]

        return to_return

    def filter(self, filter_func):
        """
        This function returns a new Storage instance containing RBPs based on
        the filter function passed. The passed in function should take a RBP
        name and return True or False.

        :param filter_func: the filter-function with which the RBPs are decided
                            for keeping or discarding.

        """
        to_return = Storage()
        for k in self._rbps:
            if filter_func(k):
                to_return[k] = self._rbps[k]

        return to_return

    def sites_analysis(self, gene, bp_threshold=0):
        """
        This function returns a dictionary mapping the binding sites of an input
        gene to a Storage instance that stores RBPs that bind within a
        threshold range of the site. The Storage variable only contains binding
        site information for the sites that are nearby to the key site and
        excludes data about other binding sites.

        :param gene: input gene (RBP) of interest
        :param bp_threshold:  (Default value = 0)

        e.g.

        RNA         : < ------------------------------------------->
        RBP1 sites  :       <--> (1)   <---> (2)       <--> (3)
        RBP2 sites  :   <-->            <------>      <->
        RBP3 sites  : <-->        <--->       <-->
        RBP4 sites  :   <-->  <->          <------>      <->


        sites_analysis(RBP1, ...) returns:

        {
            site1: Storage( ... all sites close to site 1 by any RBP ... )
            site2: Storage( ... all sites close to site 2 by any RBP ... )
            site3: Storage( ... all sites close to site 3 by any RBP ... )
        }

        """
        # sanity check
        if gene not in self._rbps:
            gene = self.synonym_func(gene)
            if gene not in self._rbps:
                raise KeyError("RBP not found in Storage")

        to_return_dict = {}  # rbp keys mapping to nearest site and distance

        for site in self[gene]:
            to_return_dict[site] = self.all_sites_in(
                site, bp_threshold=bp_threshold
            )

        return to_return_dict

    def print(self):
        """
        prints the Storage instance (collection of RBP binding sites)
        """
        print(self)

    def len(self):
        """
        returns the number of RBPs binding to this Storage (RNA) instance
        """
        return len(self)

    def all_sites_in(self, interval_range, bp_threshold=0):
        """
        returns a new Storage instance only containing binding sites contained
        within an interval range (with some leeway, if needed)
        :param interval_range: in the form (start, end) or (start, end, _)
        :param bp_threshold: number of bases away sites are allowed to be from
                             the interval_range to be included in the returned
                             Storage instance
                             (Default value = 0)

        """
        filtered_storage = Storage()
        for rbp, binding_sites in self._rbps.items():
            filtered_rbp = binding_sites.filter_overlap(
                interval_range, bp_threshold=bp_threshold
            )
            if len(filtered_rbp) > 0:
                filtered_storage[rbp] = filtered_rbp
        return filtered_storage

    def print_bed(
        self,
        chr_n=1,
        displacement=0,
        end_inclusion=False,
        add_annotation=False,
        include_score=False,
        score_max=1000,
        score_base=1000,
        include_color=False,
        conditional_color_func=-1,
        include_header=False,
        is_bar=False,
        is_additional_columns=False,
        annotation_to_additional_columns=None,
    ):
        """
        Prints the BED files representing the binding sites of all RBPs
        contained within the Storage instance (RNA molecule).

        :param chrN: the chromosome number on which the RNA is on the genome.
                     (Default value = 1)
        :param displacement: the starting base number for the RNA on the
                             chromosome  (Default value = 0)
        :param endInclusion: set to True if the formatting requires last base to
                             be included in the interval specification
                             (Default value = False)
        :param addAnnotation: set to True if the binding site annotations should
                              be included in the BED file content returned
                              (Default value = False)
        :param includeScore:  set to True if the scores should be included
                              (Default value = False)
        :param scoreMax:  (Default value = 1000)
        :param scoreBase:  (Default value = 1000)
        :param includeColor: set to True if the binding sites should be colored
                             (Default value = False)
        :param conditionalColor_func: if provided, colors the sites based on
                                      the provided function's evaluation of
                                      binding sites. Thus, the function should
                                      be of the form f: site -> color.
                                      -1 if not provided.
                                      (Default value = -1)
        :param includeHeader: if True, prints header line of BED as well.
                              (Default value = False)
        :param isBar: no idea (Default value = False)
        :param is_additional_columns: no idea (Default value = False)
        :param annotation_to_additional_columns: no idea (Default value = None)

        """
        output_str = ""
        for rbp, binding_sites in self._rbps.items():
            output_str += binding_sites.print_bed(
                name=rbp,
                chr_n=chr_n,
                displacement=displacement,
                end_inclusion=end_inclusion,
                add_annotation=add_annotation,
                include_score=include_score,
                score_max=score_max,
                score_base=score_base,
                include_color=include_color,
                conditional_color_func=conditional_color_func,
                is_bar=is_bar,
                is_additional_columns=is_additional_columns,
                annotation_to_additional_columns=annotation_to_additional_columns,
            )

        rbps = list(self.get_rbps())
        num_rbps = min(len(rbps), 3)
        rbps = rbps[:num_rbps]

        if include_header:
            header = (
                'track name="'
                + "-".join(rbps)
                + '" description="A list of binding sites of various RBPs,'
                + " including "
                + ",".join(rbps)
                + '"'
                + (' itemRgb="On"' if include_color else "")
                + (' useScore="1"' if include_score else "")
                + "\n"
            )
        else:
            header = ""
        return header + output_str

    def sum_over_all(self):
        """
        Creates and returns a new binding site object with all of the binding
        sites across all RBPs stored in the current storage. Overlap mode is set
        to True for the returned binding site object.

        """

        new_binding_site = BindingSites(overlap_mode=True)
        for rbp in self:
            for site in self[rbp]:
                # print(site)
                new_binding_site.add(site)
        return new_binding_site

    def print_wig(
        self,
        chr_no=1,
        displacement=0,
        include_name=False,
        include_description=False,
        name="",
        description="",
        include_header=True,
    ):
        """
        Prints a WIG file represeting the density of binding by the RBPs on the
        RNA molecule (Storage) instance.

        :param chr_no: chromosome number on which the RNA molecule lies
                             (Default value = 1)
        :param displacement: base number on which the RNA lies on the chromosome
                             (Default value = 0)
        :param include_name: no idea(Default value = False)
        :param include_description: no idea (Default value = False)
        :param name:  (Default value = "")
        :param description:  (Default value = "")
        :param include_header:  (Default value = True)

        """

        return self.sum_over_all().print_wig(
            chr_no=chr_no,
            displacement=displacement,
            include_name=include_name,
            include_description=include_description,
            name=name,
            description=description,
            include_header=include_header,
        )
